@@ -12,15 +12,14 @@ use Sys::Hostname;
 # [LH] 2013-08-04:  Added Fcntl for better pidfile locking.  [RT81914]
 use Fcntl qw(:DEFAULT :flock);
 
-use Error qw(:try);
-
 use Helios;
 use Helios::Error;
 use Helios::LogEntry::Levels qw(:all);
-use Helios::TheSchwartz;
+# [LH] [2013-10-04]: Switched to using Helios::TS for virtual jobtypes.
+use Helios::TS;
 use Helios::Config;
 
-our $VERSION = '2.61';
+our $VERSION = '2.80';
 
 # FILE CHANGE HISTORY
 # [2012-01-08]: Added a check to try to prevent loading code outside of @INC.
@@ -65,6 +64,28 @@ our $VERSION = '2.61';
 # Replaced write_pid_file() and mostly rewrote running_process_check() to 
 # prevent a PID file race condition.  
 # [LH] [2013-08-19]: Added comments for clarification of fix for [RT81914].
+# [LH] [2013-09-21]: Added code to enable job prioritization features in 
+# Helios::TheSchwartz.  Added code to implement WORKER_LAUNCH_PATTERN feature.
+# [LH] [2013-10-04]: Added code to implement "virtual jobtypes."  Switched to 
+# using Helios::TS instead of Helios::TheSchwartz.  Changed command line 
+# option parsing to accomodate options other than "--clear-halt".  Removed 
+# old commented out code for class name sanity checking and class loading.  
+# Removed old code to calculate workers to launch, also removed code for 
+# WORKER_BLITZ_FACTOR feature (WORKER_LAUNCH_PATTERN replaces it).
+# [LH] [2013-10-18]: Replaced most try {} blocks with eval {}.  Changed option 
+# handling to better handle --help and fix --version.  Added --debug option.  
+# Changed $CLASS to $OPT_CLASS like all the other $OPT_* option variables.  
+# Added startup message so user knows which collective db the daemon connected 
+# to.  Partially updated POD with new options and mentioned new 
+# Helios::Configuration POD; also added mention of zero_sleep_interval option.
+# [LH] [2013-11-24]: Removed use of Error module as all use of try {} catch {} 
+# blocks have been removed.  Changed command line option handling so jobtypes 
+# can be specified with multiple --jobtype options as well as a single comma-
+# delimited --jobtypes argument.  Removed old commented out code.  Database
+# reconnect code now dumps all cached database connections before attempting 
+# a reconnect.
+# [LH] [2014-02-28] POD updates.  Updated copyright info.  Changed 
+# WORKER_LAUNCH_PATTERN values to "linear","dynamic", and "optimistic".
 
 =head1 NAME
 
@@ -72,29 +93,84 @@ helios.pl - Launch a daemon to service jobs in the Helios job processing system
 
 =head1 SYNOPSIS
 
+ # Make sure HELIOS_INI is set and exported.  Optionally enable debug mode.
  export HELIOS_INI=/path/to/helios.ini
  [export HELIOS_DEBUG=1]
- helios.pl <jobclass> [--clear-halt]
+ # Full command line options 
+ helios.pl [--service=<service class>] [--jobtypes=<jobtypename,jobtypename>] \
+ [--clear-halt] [--version] [--help]
 
- # just prints version info
+ # Simple cmd line example: start a daemon with the Helios::TestService service.
+ helios.pl Helios::TestService
+ 
+ # More complex: start a Helios::TestService daemon, but have it run MyService jobs.
+ helios.pl --service=Helios::TestService --jobtypes=MyService
+
+ # Just prints version info.
  helios.pl --version
+ 
+ # Prints this help page.
+ helios.pl --help
 
 =head1 DESCRIPTION
 
-The helios.pl program, given a subclass of Helios::Service, will launch a daemon to service Helios 
-jobs of that subclass.  The number of worker processes to run concurrently and various other 
-parameters are set via a helios.ini file and the Helios database (the connection information 
-of which is also defined in helios.ini).
+The helios.pl program, given a Helios service class, will launch a daemon
+to service Helios jobs of that class.  The number of worker processes to run 
+concurrently and various other parameters are set via a helios.ini file and the 
+Helios collective database (the connection information of which is also defined 
+in helios.ini).
 
-Under normal operation, helios.pl will attempt to load the service class specified on the command 
-line and use it to read the contents of the helios.ini file.  If successful, it will attempt to 
-connect to the Helios collective database and read the relevant configuration parameters from 
-there.  If that is successful, helios.pl will daemonize and start servicing jobs of the specified 
-class.
+Under normal operation, helios.pl will attempt to load the service class 
+specified on the command line and read the contents of the helios.ini file.  If
+successful, it will attempt to connect to the Helios collective 
+database specified in helios.ini and read the configuration parameters relevant
+to the loaded service class from there.  If that is successful, helios.pl will
+then daemonize and start servicing jobs of the specified class.  If additional
+jobtypes are specified with the --jobtypes option, jobs of those additional
+types will also be serviced by the loaded service class.
 
-In debug mode (set HELIOS_DEBUG=1), helios.pl will not disconnect from the terminal, and will 
-output extra debugging information to the screen and the Helios log.  It will also enable debug mode 
-on the service class, which may also support the output of extra debugging information.
+=head1 helios.pl COMMAND LINE OPTIONS
+
+=head2 --service [REQUIRED]
+
+The --service option specifies the name of the Helios service to load and
+launch worker processes for.  If you specify the name as the first option on
+the command line, you can actually drop the '--service=' part, as helios.pl
+will assume the first option is the service class name.
+
+Examples:
+
+ helios.pl --service=MyService
+ --OR--
+ helios.pl MyService
+
+=head2 --jobtypes
+
+Normally, a Helios service will run jobs belonging to its own jobtype (the
+service class name matches the jobtype name).  However, as of Helios 2.80, a
+Helios service can run jobs of multiple jobtypes if necessary.  These
+additional jobtypes should be specified on the helios.pl command line with the
+--jobtypes option.  You can specify multiple jobtypes by separating them with
+commas, or specify multiple --jobtype values.  (Thanks to helios.pl's use of
+L<Getopt::Long>, '--jobtype' and '--jobtypes' are interchangable.)
+
+Examples:
+
+ # MyService handles jobs of jobtype "MyService" (the default)
+ helios.pl MyService
+ 
+ # MyService handles MyService jobs and MyIndexer jobs
+ helios.pl MyService --jobtype=MyIndexer
+ 
+ # MyService handles MyService, MyIndexer, and MyMailer jobs
+ helios.pl MyService --jobtypes=MyIndexer,MyMailer
+ --OR--
+ helios.pl MyService --jobtype=MyIndexer --jobtype=MyMailer
+
+For more information about jobtypes and how they relate to jobs and services,
+see the L<Helios::JobType> man page.
+ 
+=head2 --clear-halt
 
 If the --clear-halt option is specified, helios.pl will attempt to remove a HALT
 parameter specified in the Helios configuration for the specified service on the
@@ -102,12 +178,45 @@ current host.  This is helpful if you shutdown your Helios service daemon using
 the Helios::Panoptes Collective Admin view.  Note that it will NOT remove a HALT
 specified globally (where host = '*').
 
+=head2 --version
+
+The --version option displays the Helios framework version and the helios.pl
+version and then exits.
+
+=head2 --help
+
+The --help option displays this documentation.
+
+=head1 helios.pl ENVIRONMENT VARIABLES
+
+The Helios framework relies directly on two environment variables to function
+correctly.  Both should be exported as well as set to ensure the helios.pl
+child processes can see them as well as the main helios.pl daemon.
+
+=head2 HELIOS_INI [REQUIRED]
+
+The full path to the helios.ini configuration file.  This should be an absolute
+path, not a relative one.  The default is to use a helios.ini in the current
+directory if it exists, which will work if you are running in Debug Mode (see
+HELIOS_DEBUG below), but will cause problems if helios.pl is running normally.
+
+I<Default:  ./helios.ini>
+
+=head2 HELIOS_DEBUG
+
+Setting HELIOS_DEBUG to 1 causes helios.pl to run in Debug Mode.  In Debug Mode,
+helios.pl will not disconnect from the terminal and will output extra debugging
+information to the screen and the Helios log.  It will also enable debug mode
+on the service class, which may support the output of extra debugging
+information if you so choose.
+
+I<Default:  undef>
+
 =head1 HELIOS.INI
 
-The initial parameters for helios.pl are defined in an INI-style configuration file typically 
-named "helios.ini".  The file's location is normally specified by setting the HELIOS_INI 
-environment variable before helios.pl is started.  If HELIOS_INI is not set, helios.pl will default 
-to the helios.ini in the current directory, if present.
+The initial parameters for helios.pl are defined in an INI-style configuration 
+file typically named "helios.ini".  The file's location is normally specified 
+by setting the HELIOS_INI environment variable before helios.pl is started.  
 
 A helios.ini file normally contains a [global] section with the parameters 
 necessary to connect to the Helios collective database, and any parameters local
@@ -126,132 +235,112 @@ Example helios.ini:
  master_launch_interval=60
  zero_launch_interval=90
 
- [SearchIndex::LoadTestService]
+ [AnotherService]
  OVERDRIVE=1
  MAX_WORKERS=3
 
 =head2 Helios.ini Parameters for helios.pl
 
+This section just covers the basic config options for helios.pl.  For a 
+complete list of built-in Helios config options, consult the 
+L<Helios::Configuration> man page.
+
 =head3 Configuration options to place in the [global] section:
 
+These options will be the same for all Helios services running a host that 
+share the same collective database, so these options must be placed in a 
+helios.ini section named [global] so they are visible to all running helios.pl 
+instances.
+
 =over 4
 
-=item dsn
+=item dsn [REQUIRED]
 
-Datasource name for the Helios parameter database.  This will also contain the data structures for 
-the underlying TheSchwartz queuing system.
+Datasource name for the Helios collective database.  The collective database
+houses the data structures for service configuration, jobtypes, job queuing and
+history, and the default logging subsystem.
 
-=item user
+I<Default: none>
+
+=item user [REQUIRED]
 
 Database user for the datasource name described above.
 
-=item password
+I<Default: none>
+
+=item password [REQUIRED]
 
 Database password for the datasource name described above.
 
+I<Default: none>
+
 =item pid_path
 
 The location where helios.pl should write the PID file for this service 
-instance.  The default is "/var/run/helios".  The name of the PID file will be 
-a variation on the service class's name.
+instance.  The name of the PID file will be a variation on the service class's
+name.
+
+I<Default:  /var/run/helios>
 
 =back
 
 =head3 Configuration options to place in individual service sections:
 
+The options listed in this section are available to tune helios.pl to work 
+better with your Helios service class and the jobs it needs to service.  These
+are read at startup and, unlike some other Helios config options, cannot be 
+dynamically changed while helios.pl is running.  If you wish to tune one of 
+these parameters, reset the parameter and restart the service daemon.
+
 =over 4
 
 =item master_launch_interval
 
-Set the master_launch_interval to determine how long helios.pl should sleep after launching 
-workers before accessing the database to update its configuration parameters and check for waiting 
-jobs.  The default is 1 second which works well for normally short-lived jobs (2 secs or less), 
-but that may be overkill for longer-lived jobs (jobs that run longer than 2 sec).  Setting the 
-master_launch_interval option may help prevent needless database traffic for longer running jobs.
+Set the master_launch_interval to determine how long helios.pl should sleep 
+after launching workers before accessing the database to update its 
+configuration parameters and check for waiting jobs.  The default is 1 second, 
+which should be sufficient for most applications. 
+
+I<Default:  1>
 
 =item zero_launch_interval
 
-Set the zero_launch_interval to determine how long helios.pl should sleep after reaching its 
-MAX_WORKERS limit.  The default is 10 sec.  If jobs are running long enough that 
-helios.pl is frequently hitting is MAX_WORKERS limit (there are waiting jobs but 
-helios.pl can't launch new workers because previously launched jobs are still running), increasing 
-the zero_launch_interval will reduce needless database traffic.
+Set the zero_launch_interval to determine how long helios.pl should sleep after 
+reaching its MAX_WORKERS limit.  The default is 10 sec.  If jobs are running 
+long enough that helios.pl is frequently hitting the MAX_WORKERS limit (there 
+are waiting jobs but helios.pl can't launch new workers because previously 
+launched jobs are still running), increasing the zero_launch_interval will 
+reduce needless database traffic.
+
+I<Default:  10>
+
+=item zero_sleep_interval
+
+Set the zero_sleep_interval to adjust the amount of time between checks for 
+available jobs in the job queue when the job queue is empty.  If the helios.pl 
+daemon determines there are no available jobs for a service, it sleeps 
+zero_sleep_interval seconds and then checks for jobs again.  The default is 10 
+sec.  If you notice jobs are sitting in the job queue too long before workers 
+are launched to service them, reduce this number to cause jobs to be started 
+faster.  If you have a small number of jobs and do not care if they sit in the 
+job queue for a few seconds before being serviced, increase this number to 
+reduce database queries. 
+
+I<Default:  10>
 
 =back
 
-=head1 HELIOS CTRL PANEL (helios_params_tb)
+=head1 HELIOS_CONFIG_* AND HELIOS CTRL PANEL (helios_params_tb)
 
-In addition to helios.ini, certain helios.pl configuration options can be set via the Ctrl Panel 
-in the Helios::Panoptes web interface.  These configuration options are read by helios.pl from the 
-HELIOS_PARAMS_TB table in the Helios database.  Though the Ctrl Panel is designed mostly to 
-provide a standardized interface for service classes to store and retrieve configuration 
-parameters, certain helios.pl control parameters can be set here as well.  
-
-The helios.pl daemon will refresh its configuration parameters from the database after waiting 
-master_launch_interval seconds after launching worker processes.  If MAX_WORKERS
-workers are still running, the helios.pl daemon will wait zero_launch_interval seconds to refresh
-its configuration parameters and try to launch workers again.  If you have long running jobs (jobs 
-that last more than a few seconds), resetting these parameters higher may help smooth processing 
-and reduce needless database traffic.  See the HELIOS OPERATION section below for more 
-information.
-
-=head2 helios.pl Control Parameters in Ctrl Panel
-
-=over 4
-
-=item MAX_WORKERS
-
-The Helios system is designed to run one daemon process for each service class per host.  However, 
-each service daemon can run more than 1 worker process simultaneously.  The upper limit for worker 
-processes is set using the MAX_WORKERS option.  The helios.pl default is 1 worker per 
-server.  You can provide a different default for a particular class by specifying a value in 
-helios.ini.  If active management of available workers is not necessary for a 
-particular service class, this option can be set in helios.ini.
-
-=item OVERDRIVE
-
-Speed up job processing by allowing workers to service jobs until no more are in the queue.  
-Normally, a worker process will service a single job and then exit (1:1 worker/job ratio).  With 
-OVERDRIVE mode enabled, a worker process will keep servicing jobs until it can find no more of 
-that type in the job queue (1:many worker/job ratio).  This may help to speed up processing for 
-short-lived jobs, as once a worker process has started, it will stay in memory, continuing to 
-process jobs until there are no more to process.  This can also enable on-the-fly caching of 
-resources, as once a worker has loaded a resource or established a database connection for a job 
-it can be held for use by subsequent jobs, potentially reducing database load or disk I/O.  Longer 
-running jobs may not benefit as much, however.
-
-Like MAX_WORKERS, if active management of the job processing mode is not necessary, this
-parameter can be set in helios.ini (though that is discouraged).
-
-=item HOLD
-
-If set to 1, a HOLD parameter will cause the helios.pl service daemon to stop launching new workers 
-to process jobs.  Currently processing jobs will be allowed to complete, and those worker 
-processes will end normally.    
-
-Setting HOLD to 0 will allow normal job processing to resume.
-
-=item HALT
-
-If a HALT parameter is defined, the helios.pl service daemon will start process cleanup and will 
-exit.  As with HOLD = 1, jobs currently processing will complete, and the worker processes 
-will exit normally.  
-
-Once a helios.pl daemon has HALTED, there is no way to restart it from the Helios::Pantoptes 
-interface.  It must be restarted from the command line with a 'helios.pl <serviceclass>' command.
-
-As of Helios 2.22, service daemons can also be shut down in the normal *nix manner of using 
-the 'kill' command to send the daemon a SIGTERM signal.
-
-=item WORKER_MAX_TTL
-
-The time in seconds to allow a worker to run.  If you have a problem with 
-workers that get stuck while performing a job (perhaps waiting for a database
-connection, or some other problem), you can set WORKER_MAX_TTL for that service.
-The helios.pl program will check to see if workers for that service are 
-running beyond their intended time-to-live, and kill them if they run too long.
-
-=back
+In addition to helios.ini, certain helios.pl configuration options can be set 
+via the helios_config_* commands or the Ctrl Panel or Collective Admin views
+in the Helios::Panoptes web admin interface.  These configuration options are
+read by helios.pl from the HELIOS_PARAMS_TB table in the Helios collective
+database.  For more information on configuration parameters available to tune
+how a service runs jobs, see the L<Helios::Configuration> man page.  For more
+information on the L<helios_config_get>, L<helios_config_set>,
+L<helios_config_unset> commands and the L<Helios::Panoptes> web admin
+interface, see their respective man pages or Perl POD.
 
 =cut
 
@@ -263,7 +352,32 @@ our $HELIOS_PL_PATH = File::Spec->catfile($FindBin::Bin, $basename);
 # globals settings
 our $DEBUG_MODE = $ENV{HELIOS_DEBUG};
 our $HELIOS_INI = $ENV{HELIOS_INI};
-our $CLASS = shift @ARGV;
+# BEGIN CODE Copyright (C) 2013 by Logical Helion, LLC.
+# [LH] [2013-10-04]: Virtual jobtypes.  We have to parse the cmd line args 
+# in a completely different way, but still maintain backward compatibility. 
+# if the first cmd line switch starts with '-', then all the options are 
+# "proper."  If not, then the first option is the service class, and the 
+# remaining options will be proper.
+our $OPT_CLASS      = '';
+our $OPT_CLEAR_HALT = 0;
+our $OPT_DEBUG      = 0;
+our $OPT_HELP       = 0;
+our @OPT_JOBTYPES   = ();
+our $OPT_VERSION    = 0;
+if ( defined($ARGV[0]) && $ARGV[0] !~ /^\-/) {
+	$OPT_CLASS = shift @ARGV;
+}
+GetOptions(
+	"service=s"  => \$OPT_CLASS,
+	"clear-halt" => \$OPT_CLEAR_HALT,
+	"help"       => \$OPT_HELP,
+	"jobtypes=s" => \@OPT_JOBTYPES,
+	"version"    => \$OPT_VERSION,
+	"debug"      => \$OPT_DEBUG,
+);
+our @ALT_JOBTYPES = split(/,/, join(','=>@OPT_JOBTYPES));
+$DEBUG_MODE = 1 if $OPT_DEBUG;
+# END CODE Copyright (C) 2013 by Logical Helion, LLC.
 
 # other globals
 # max workers will default to 1 if not set elsewhere
@@ -276,7 +390,9 @@ our %DEFAULTS = (
     ZERO_SLEEP_INTERVAL => 10,
     REGISTRATION_INTERVAL => 60,
     WORKER_BLITZ_FACTOR => 1,
-    WORKER_MAX_TTL_WAIT_INTERVAL => 20
+    WORKER_MAX_TTL_WAIT_INTERVAL => 20,
+    PRIORITIZE_JOBS => 0,
+	WORKER_LAUNCH_PATTERN => 'linear',
 );
 our $CLEAN_SHUTDOWN = 1;				# used to determine if we should remove the PID file or not (at least for now)
 
@@ -309,30 +425,27 @@ our $WORKER_BLITZ_FACTOR = 1;			# used to determine how many workers to launch
 our $HELIOS_DB_CONN;					# database handle to help with DBI+forking issues
 
 # print help if asked
-if ( !defined($CLASS) || ($CLASS eq '--help') || ($CLASS eq '-h') ) {
+# [LH] [2013-10-04]: New @ARGV parsing for "--help". 
+if ( $OPT_HELP ) {
 	require Pod::Usage;
 	Pod::Usage::pod2usage(-verbose => 2, -exitstatus => 0);
 }
 
 # conditionally load module or die in the attempt
-my $worker_class = $CLASS;
+my $worker_class = $OPT_CLASS;
 print "Helios ",$Helios::VERSION,"\n";
 print "helios.pl Service Daemon version $VERSION\n";
 # --version support
-if ( $CLASS eq '--version') { exit(); }
-#if ( $worker_class !~ /^[A-Za-z]([A-Za-z0-9_\-]|:{2})*[A-Za-z0-9_\-]$/ ) {
-#	print "Sorry, requested name is invalid: $worker_class\n";
-#	exit(1);
-#}
-print "Attempting to load $worker_class...\n"; 
-#unless ( $worker_class->can('new') ) {
-#        eval "require $worker_class";
-#        die $@ if $@;
-#}
-unless ( $worker_class ) {
-	print "You must specify the module name of a Helios service.\n";
+# BEGIN CODE Copyright (C) 2013 by Logical Helion, LLC.
+# [LH] [2013-10-04]: New @ARGV parsing for "--version".
+# [LH] [2013-10-18]: New @ARGV parsing for --service.
+if ( $OPT_VERSION ) { exit(0); }
+unless ($OPT_CLASS) { 
+	warn("The name of a service class is required.\n");
 	exit(1);
 }
+# END CODE Copyright (C) 2013 by Logical Helion, LLC.
+print "Attempting to load $worker_class...\n"; 
 require_module($worker_class);
 
 if ( defined($worker_class->VERSION) ) {
@@ -351,6 +464,14 @@ if ( defined($HELIOS_INI) ) {
         $worker->setIniFile( $DEFAULTS{HELIOS_INI} );
 }
 $worker->setJobType($worker_class);
+# BEGIN CODE Copyright (C) 2013 by Logical Helion, LLC.
+# [LH] [2013-10-04]: Virtual jobtypes set & init.
+if (@ALT_JOBTYPES) {
+	$worker->setAltJobTypes(@ALT_JOBTYPES); 
+	$worker->lookupAltJobtypeids();
+	print "Servicing jobtypes: ", join(' ' => ($worker->getJobType(), $worker->getAltJobTypes())),"\n";
+}
+# END CODE Copyright (C) 2013 by Logical Helion, LLC.
 $worker->debug($DEBUG_MODE);
 eval {
 	$worker->prep();
@@ -402,6 +523,9 @@ if ( defined($params->{WORKER_MAX_TTL_WAIT_INTERVAL}) ) {
 # make a globally accessible database handle 
 # to make it easier to clean up CachedKids after a fork()
 $HELIOS_DB_CONN = $worker->dbConnect();
+# [LH] [2013-10-18]: Added startup message so user knows which collective db
+# we connected to
+print "Connected to collective database: ",$params->{dsn},"\n";
 
 if ($DEBUG_MODE) { 
 	print "MASTER LAUNCH INTERVAL: $MASTER_LAUNCH_INTERVAL\n"; 
@@ -433,7 +557,8 @@ my $times_sleeping = 0;
 # then we'll check to see if HALT is still there
 # if it is, we'll have to stop here rather than 
 # daemonize and launch the main loop
-if ( defined($ARGV[0]) && lc($ARGV[0]) eq '--clear-halt' ) {
+# [LH] [2013-10-04]: New @ARGV parsing for "--clear-halt". 
+if ( $OPT_CLEAR_HALT ) {
 	clear_halt();
 	$worker->prep();
 	$params = $worker->getConfig();
@@ -465,6 +590,7 @@ unless ($DEBUG_MODE) {
 $SIG{CHLD} = \&reaper;
 $SIG{TERM} = \&terminator;
 
+
 =head1 HELIOS OPERATION
 
 After initial setup, the helios.pl daemon will enter a main operation loop where configuration 
@@ -482,8 +608,8 @@ Refresh configuration parameters from database.
 
 =item 2.
 
-Check job queue to see if there are jobs available for processing (if not, sleep for 
-zero_sleep_interval seconds and start again).
+Check job queue to see if there are jobs of the correct jobtype(s) available
+for processing (if not, sleep for zero_sleep_interval seconds and start again).
 
 =item 3.
 
@@ -494,11 +620,11 @@ the running jobs to complete (the default is 10 secs).
 
 =item 4.
 
-Subtract the number of currently running workers from MAX_WORKERS and launch that 
-many workers to handle the available jobs.  If MAX_WORKERS is 5 and only 2 
-workers are running, that means the helios.pl daemon will launch 3 workers.  If 
-there are less than MAX_WORKERS jobs available for processing, only 1 worker 
-will be launched to reduce job contention between workers in the collective.
+If there are jobs available and the MAX_WORKERS limit has not been reached,
+determine the number of additional workers to launch and launch them.  The
+number of workers to launch is governed by the WORKER_LAUNCH_PATTERN
+configuration parameter; see the L<Helios::Configuration> man page for more
+information.
 
 =item 5.
 
@@ -509,7 +635,8 @@ Sleep master_launch_interval seconds and start the operation loop again.
 =cut
 
 MAIN_LOOP:{
-	try {
+	# [LH] [2013-11-24]: Replaced try {} otherwise {} block with eval {} or do {}. 
+	eval {
 	
 		# while not halted
 		while (!defined($params->{HALT}) ) {
@@ -584,14 +711,8 @@ MAIN_LOOP:{
 				# LAUNCHING WORKERS
 				# if we've got to this part of the loop, we have jobs that need workers to launch
 				# (though we still may not do it if we've already reached our limit)
-				my $workers_to_launch = $max_workers - $running_workers;
-				# if the number of waiting jobs is less than (max workers * worker blitz factor)
-				# then only launch one worker to reduce worker contention
-				# refresh worker blitz factor, then determine if we blitz or not
-				$WORKER_BLITZ_FACTOR = $params->{WORKER_BLITZ_FACTOR} ? $params->{WORKER_BLITZ_FACTOR} : $DEFAULTS{WORKER_BLITZ_FACTOR};
-				if ( ($workers_to_launch > 0) && ($waiting_jobs < ($max_workers * $WORKER_BLITZ_FACTOR ) ) ) {
-					$workers_to_launch = 1;
-				}
+				# [LH] [2013-09-21]: Added code to implement WORKER_LAUNCH_PATTERN feature.
+				my $workers_to_launch = workers_to_launch($waiting_jobs, $running_workers, $max_workers);				
 				$worker->logMsg(LOG_NOTICE, "$waiting_jobs jobs waiting; $running_workers workers running; launching $workers_to_launch workers");
 				for (my $i = 0; $i < $workers_to_launch; $i++) {
 		
@@ -663,8 +784,9 @@ MAIN_LOOP:{
 					sleep $MASTER_LAUNCH_INTERVAL;
 				}
 		}
-		
-	} otherwise {
+		# [LH] [2013-11-24]: Replaced try {} otherwise {} block with eval {} or do {}. 
+		1;
+	} or do {
 		my $e = shift;
 		# if we're a worker process and we ended up here
 		# it's a fluke caused by the database instability
@@ -675,13 +797,24 @@ MAIN_LOOP:{
 			print "EXCEPTION THROWN: ",$e->text,"\n"; 
 			print "ATTEMPTING TO RECONNECT...\n";
 		}
+# BEGIN CODE Copyright (C) 2013 by Logical Helion, LLC.
+		#[] dump all the database connections
+		# [LH] [2013-11-24]: Dump all the cached db connections so we can make
+		# sure we're starting over.
+		my $ck = $HELIOS_DB_CONN->{Driver}->{CachedKids};
+		%$ck = ();
+		$HELIOS_DB_CONN->disconnect() if ( ref($HELIOS_DB_CONN) && $HELIOS_DB_CONN->isa('DBI'));
+# END CODE Copyright (C) 2013 by Logical Helion, LLC.
+
 		my $retry;
 		my $return_code = 0;
 		for($retry = 1; $retry <= $SAFE_MODE_RETRIES; $retry++) {
 			my $success = 0;
-			try {
+			# [LH] [2013-11-24]: Replaced try {} otherwise {} block with eval {} or do {}. 
+			eval {
 				$success = $worker->dbConnect();
-			} otherwise {
+				1;
+			} or do {
 				# actually, if we fail, we do nothing
 			};
 			if ($success) {
@@ -696,6 +829,11 @@ MAIN_LOOP:{
 			print "DATABASE RECONNECTION ATTEMPTS FAILED!\n";
 		}
 		if ($return_code) {
+# BEGIN CODE Copyright (C) 2013 by Logical Helion, LLC.
+			# [LH] [2013-11-24]: Reconnect the main db connection variable before
+			# we go back to the main loop.
+			$HELIOS_DB_CONN = $worker->dbConnect();
+# END CODE Copyright (C) 2013 by Logical Helion, LLC.
 			$worker->logMsg(LOG_CRIT, "Exiting SAFE MODE; Reestablished connection to database after exception: ".$e->text);
 			$worker->errstr(undef);
 			redo MAIN_LOOP;
@@ -727,6 +865,11 @@ exit();
 
 =head1 SUBROUTINES
 
+The helios.pl program calls several of its own functions at various times
+during startup, shutdown, and during main loop operation.  These are probably
+only of limited interest, but are documented here for those wanting to
+understand better the guts of the helios.pl daemon. 
+
 =head1 PROCESS CONTROL FUNCTIONS
 
 =head2 daemonize() 
@@ -774,7 +917,6 @@ sub daemonize {
 	# session and making our own process group
 	# as well as closing any standard open filehandles.
 	POSIX::setsid();
-#[] I may have changed religions on this one
 #	close (STDIN); 
 #	close (STDOUT); 
 #	close (STDERR);
@@ -818,11 +960,11 @@ sub double_clutch {
 
 The launch_worker() function launches a new worker process.  
 After the fork() from the main process, the new child process will call 
-launch_worker().  The launch_worker() function will instantiate a new 
-TheSchwartz object, set the object's B<ability> (TheSchwartz term) to the class 
-of the loaded service, and starts the worker on its way by calling the work() 
+launch_worker() to instantiate and configure a new L<Helios::TS> job queuing
+system object and start the worker on its way by calling the object's work() 
 method.  If the OVERDRIVE run mode is enabled for the service, the 
-work_until_done() method is called instead.
+work_until_done() method is called instead.  See L<Helios::Configuration> for
+information about running workers in OVERDRIVE vs. Normal Mode.
 
 =cut
 
@@ -830,8 +972,27 @@ sub launch_worker {
     # just in case this would cause a problem in worker process
     $SIG{CHLD} = 'DEFAULT';
     $SIG{TERM} = 'DEFAULT';
-	my $client = Helios::TheSchwartz->new(databases => $DATABASES_INFO);
+ 
+# BEGIN CODE Copyright (C) 2012-3 by Logical Helion, LLC.
+	# [LH] [2013-09-21]: Added code to enable job prioritization features in Helios::TheSchwartz.
+	# [LH] [2013-10-04]: Virtual jobtypes:  switched to using Helios::TS.
+	my Helios::TS $client = Helios::TS->new(
+		databases  => $DATABASES_INFO,
+		prioritize => defined($params->{PRIORITIZE_JOBS}) ? $params->{PRIORITIZE_JOBS} : $DEFAULTS{PRIORITIZE_JOBS}
+	);
+# END CODE Copyright (C) 2012-3 by Logical Helion, LLC.
 	$client->can_do($worker_class);
+# BEGIN CODE Copyright (C) 2013 by Logical Helion, LLC.
+	# [LH] [2013-10-04]: Virtual jobtypes:  Set up Helios::TS object with 
+	# alt jobtypes and designate the "Active Worker Class" (the service class
+	# that is actually running the jobs).
+	if ($worker->getAltJobTypes) {
+		$client->set_active_worker_class($worker_class);
+		for ($worker->getAltJobTypes) {
+			$client->can_do($_);
+		}
+	}
+# END CODE Copyright (C) 2013 by Logical Helion, LLC.
 	my $return;
 	if ( defined($params->{OVERDRIVE}) && $params->{OVERDRIVE} == 1 ) {
 		$return = $client->work_until_done();			
@@ -932,12 +1093,12 @@ sub terminator {
     my $TERM = 0;
     # did we receive a SIGTERM, or was the HALT config param set?
     if ( defined($params->{HALT}) ) {
-        $worker->logMsg(LOG_NOTICE, "HALTING $CLASS on host ".$worker->getHostname());
+        $worker->logMsg(LOG_NOTICE, "HALTING $OPT_CLASS on host ".$worker->getHostname());
     } else {
         # tell the workers they need exit
         $TERM = 1;
-        $worker->logMsg(LOG_NOTICE, "Received TERM signal; HALTING $CLASS on host ".$worker->getHostname());
-        $worker->logMsg(LOG_NOTICE, "Setting HALT for $CLASS on host ".$worker->getHostname());
+        $worker->logMsg(LOG_NOTICE, "Received TERM signal; HALTING $OPT_CLASS on host ".$worker->getHostname());
+        $worker->logMsg(LOG_NOTICE, "Setting HALT for $OPT_CLASS on host ".$worker->getHostname());
         set_halt();
     }
 
@@ -946,7 +1107,7 @@ sub terminator {
     sleep $ZERO_LAUNCH_INTERVAL;
     sleep $ZERO_LAUNCH_INTERVAL;
     # reap any processes that ended while we were sleeping
-    $worker->logMsg(LOG_NOTICE, "Reaping $CLASS workers on host ".$worker->getHostname());
+    $worker->logMsg(LOG_NOTICE, "Reaping $OPT_CLASS workers on host ".$worker->getHostname());
     reaper();
     
     # kill any workers still running (they've had plenty of time to end on their own)
@@ -954,7 +1115,7 @@ sub terminator {
         if ( kill 0 => $pid ) {
             # it's still alive, kill it
             kill 9, $pid;
-            $worker->logMsg(LOG_ERR, "Killed process $pid (shutdown $CLASS instance)");
+            $worker->logMsg(LOG_ERR, "Killed process $pid (shutdown $OPT_CLASS instance)");
         }
         delete $workers{$pid};
     }
@@ -963,10 +1124,10 @@ sub terminator {
     # only clear the HALT if we received a SIGTERM (we set the halt ourselves)
     if ($TERM) { 
         clear_halt(); 
-        $worker->logMsg(LOG_NOTICE, "Cleared HALT for $CLASS on host ".$worker->getHostname());
+        $worker->logMsg(LOG_NOTICE, "Cleared HALT for $OPT_CLASS on host ".$worker->getHostname());
     }
     clean_shutdown();
-    $worker->logMsg(LOG_NOTICE, "$CLASS on host ".$worker->getHostname().' HALTED.');
+    $worker->logMsg(LOG_NOTICE, "$OPT_CLASS on host ".$worker->getHostname().' HALTED.');
     exit(1);
 }
 
@@ -1130,13 +1291,15 @@ job processing.)
 =cut
 
 sub register {
-	try {
+	# [LH] [2013-10-18]: Replaced try {} block with eval {}.
+	eval {
 		my $dbh = $worker->dbConnect();
 		$dbh->do("DELETE FROM helios_worker_registry_tb WHERE worker_class = ? AND host = ?", undef, 
 					$worker->getJobType(), $worker->getHostname) or die;
 		$dbh->do("INSERT INTO helios_worker_registry_tb (register_time, start_time, worker_class, worker_version, host, process_id) VALUES (?,?,?,?,?,?)", undef,
 					time(), $START_TIME, $worker->getJobType(), $worker->VERSION, $worker->getHostname, $$) or die;
-	} otherwise {
+		1;
+	} or do {
 		throw Helios::Error::DatabaseError($DBI::errstr);
 	};
 	return 1;
@@ -1151,12 +1314,14 @@ It is called whenever there is a clean shutdown.
 =cut
 
 sub unregister {
-	try {
+	# [LH] [2013-10-18]: Replaced try {} block with eval {}.
+	eval {
 		my $dbh = $worker->dbConnect();
 		$dbh->do("DELETE FROM helios_worker_registry_tb WHERE worker_class = ? AND host = ?", undef, 
 					$worker->getJobType(), $worker->getHostname) or die;
 		$dbh->disconnect();
-	} otherwise {
+		1;
+	} or do {
 		throw Helios::Error::DatabaseError($DBI::errstr);
 	};
 	return 1;
@@ -1173,14 +1338,16 @@ when the current job is completed.
 =cut
 
 sub set_halt {
-	try {
+	# [LH] [2013-10-18]: Replaced try {} block with eval {}.
+	eval {
 		if ($DEBUG_MODE) { 
 			print "Attempting to set HALT for ",$worker->getJobType()," on ",$worker->getHostname(),"\n"; 
 		}
 		my $dbh = $worker->dbConnect();
 		$dbh->do("INSERT INTO helios_params_tb (worker_class, host, param, value) VALUES (?,?,?,?)", undef, 
 					$worker->getJobType(), $worker->getHostname, 'HALT', '1');
-	} otherwise {
+		1;
+	} or do {
 		throw Helios::Error::DatabaseError($DBI::errstr);
 	};
 	return 1;	
@@ -1197,14 +1364,16 @@ HALT parameter (where the host is specified as '*').
 =cut
 
 sub clear_halt {
-	try {
+	# [LH] [2013-10-18]: Replaced try {} block with eval {}.
+	eval {
 		if ($DEBUG_MODE) { 
 			print "Attempting to delete HALT for ",$worker->getJobType()," on ",$worker->getHostname(),"\n"; 
 		}
 		my $dbh = $worker->dbConnect();
 		$dbh->do("DELETE FROM helios_params_tb WHERE worker_class = ? AND host = ? AND param = ?", undef, 
 					$worker->getJobType(), $worker->getHostname, 'HALT');
-	} otherwise {
+		1;
+	} or do {
 		throw Helios::Error::DatabaseError($DBI::errstr);
 	};
 	return 1;	
@@ -1242,10 +1411,92 @@ sub require_module {
 
 # END CODE Copyright (C) 2012 by Logical Helion, LLC.
 
+# BEGIN CODE Copyright (C) 2012-4 by Logical Helion, LLC.
+# [LH] [2013-09-21]: Code to implement WORKER_LAUNCH_PATTERN feature.
+# [LH] [2014-02-28]: Changed algorithm names to:  linear, dynamic, optimistic.
+# Wrote documentation for function.
+
+=head2 workers_to_launch($waiting_jobs, $running_workers, $max_workers)
+
+Given the number of waiting jobs, currently running workers, and maximum
+workers to launch, workers_to_launch() returns the number of worker processes
+helios.pl should launch to correctly handle the current job workload.  There
+are several different algorithms helios.pl can use to determine this number,
+which is dependent on the environment, available resources, and type of
+application.  See L<Helios::Configuration> for the list of possible values and
+an explanation of the worker launching algorithms.
+
+=cut
+
+sub workers_to_launch {
+	my $waiting_jobs = shift;
+	my $running_workers = shift;
+	my $max_workers = shift;
+	my $wlp = defined($params->{WORKER_LAUNCH_PATTERN}) ? $params->{WORKER_LAUNCH_PATTERN} : $DEFAULTS{WORKER_LAUNCH_PATTERN};
+
+	# how many workers are we able to launch?
+	my $available_workers = $max_workers - $running_workers;
+	
+	# if we cannot launch any more workers, then we cannot launch any workers
+	if ( $available_workers <= 0 ) {
+		return 0;
+	}
+
+	# if the waiting jobs >= max_workers, forget all the launcher patterns and blitz
+	if ( $waiting_jobs >= $max_workers ) {
+		return $available_workers;
+	}
+
+	# ok, we might be able to launch some workers
+	SWITCH: {
+		# optimistic
+		# for big, fast collective databases with lots of short running jobs
+		if ( $wlp eq 'optimistic') {
+			# launch a worker for each waiting job, 
+			# or all available workers
+			# whichever is less
+			if ( $available_workers < $waiting_jobs ) {
+				return $available_workers;
+			} else {
+				return $waiting_jobs;
+			}
+		}
+		# dynamic
+		# for fast collectives that need to launch workers more slowly
+		# e.g. an app with a slow or limited shared resource
+		# this will be the the Helios 3.x default
+		if ( $wlp eq 'dynamic') {
+			# launch the difference between running workers and waiting jobs
+			# e.g. 14 waiting jobs, 10 workers running, launch 4
+			# if waiting jobs exceed available workers, just launch available workers
+			my $diff = $waiting_jobs - $running_workers;
+			if ( $diff <= 0  ) { $diff = 1; }
+			if ( $available_workers < $diff ) {
+				return $available_workers;
+			} else {
+				return $diff;
+			}
+		}
+		# linear
+		# for slow collective databases, or apps that only run a 
+		# smaller number of jobs, or longer running jobs
+		# this was the Helios 2.x default
+		if ( $wlp eq 'linear') {
+			return 1;
+		}
+		# the default is actually 'linear'
+		return 1;
+	}
+
+
+}
+
+# END CODE Copyright (C) 2012-4 by Logical Helion, LLC.
+
 
 =head1 SEE ALSO
 
-L<Helios>, L<Helios::Service>
+L<Helios>, L<Helios::Tutorial>, L<Helios::Service>, L<Helios::Configuration>
 
 =head1 AUTHOR
 
@@ -1256,7 +1507,7 @@ Andrew Johnson, E<lt>lajandy at cpan dotorgE<gt>
 Copyright (C) 2007-9 by CEB Toolbox, Inc.
 
 Portions of this software, where noted, are
-Copyright (C) 2012-2013 by Logical Helion, LLC.
+Copyright (C) 2012-2014 by Logical Helion, LLC.
 
 This program is free software; you can redistribute it and/or modify
 it under the same terms as Perl itself, either Perl version 5.8.0 or,
